import os
from typing import List
import concurrent.futures
from itertools import repeat
from injectable import injectable, autowired, Autowired
from tekleo_common_message_protocol import OdSample, OdLabeledItem, PointRelative
from tekleo_common_utils import UtilsImage


@injectable
class UtilsDatasetPascalvoc:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage)):
        self.utils_image = utils_image

    def load_sample_from_png_and_pascal_voc_xml(self, image_file_path: str, xml_file_path: str) -> OdSample:
        image_pil = self.utils_image.open_image_pil(image_file_path)
        xml_file = open(xml_file_path, 'r')
        xml_text = xml_file.read()
        xml_file.close()

        name = [line for line in xml_text.split('\n') if '<filename>' in line][0].replace('<filename>', '').replace('</filename>', '').strip()
        boxes = []

        objects = xml_text.split('<object>')
        objects = objects[1:]
        for object in objects:
            lines = object.split('\n')
            line_name = [line for line in lines if '<name>' in line][0]
            line_xmin = [line for line in lines if '<xmin>' in line][0]
            line_ymin = [line for line in lines if '<ymin>' in line][0]
            line_xmax = [line for line in lines if '<xmax>' in line][0]
            line_ymax = [line for line in lines if '<ymax>' in line][0]

            label = line_name.replace('<name>', '').replace('</name>', '').strip()
            xmin = int(line_xmin.replace('<xmin>', '').replace('</xmin>', '').strip())
            ymin = int(line_ymin.replace('<ymin>', '').replace('</ymin>', '').strip())
            xmax = int(line_xmax.replace('<xmax>', '').replace('</xmax>', '').strip())
            ymax = int(line_ymax.replace('<ymax>', '').replace('</ymax>', '').strip())

            x = xmin / image_pil.width
            y = ymin / image_pil.height
            w = (xmax - xmin) / image_pil.width
            h = (ymax - ymin) / image_pil.height

            mask = [PointRelative(x, y), PointRelative(x + w, y), PointRelative(x + w, y + h), PointRelative(x, y + h)]
            box = OdLabeledItem(label, mask)
            boxes.append(box)

        return OdSample(name, image_pil, boxes)

    def load_sample_from_folder(self, image_and_xml_file_name: str, folder_path: str) -> OdSample:
        # Build image file path, trying different image format options
        image_file_path = folder_path + '/' + image_and_xml_file_name + '.png'
        if not os.path.isfile(image_file_path):
            image_file_path = image_file_path.replace('.png', '.jpeg')
            if not os.path.isfile(image_file_path):
                image_file_path = image_file_path.replace('.jpeg', '.jpg')

        # Build XML file path, and show warning if no markup found
        xml_file_path = folder_path + '/' + image_and_xml_file_name + '.xml'
        if not os.path.isfile(xml_file_path):
            print('UtilsDatasetPascalvoc.load_sample_from_folder(): Warning! XML not found, xml_file_path=' + str(xml_file_path))
            return None

        # Load sample
        return self.load_sample_from_png_and_pascal_voc_xml(image_file_path, xml_file_path)

    def load_samples_from_folder(self, folder_path: str) -> List[OdSample]:
        samples = []

        # Get all files, strip their extensions and resort
        all_files = os.listdir(folder_path)
        all_files = ['.'.join(f.split('.')[:-1]) for f in all_files]
        all_files = set(all_files)
        all_files = sorted(all_files)

        # Load samples in parallel
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=8)
        for sample in executor.map(self.load_sample_from_folder, all_files, repeat(folder_path)):
            if sample is not None:
                samples.append(sample)

        # Filter out None values
        samples = [s for s in samples if s is not None]

        return samples



import os
import datetime
import collections
import json
import uuid
import imgviz
import labelme
import numpy
import pycocotools.mask
from typing import List
from injectable import injectable, autowired, Autowired
from tekleo_common_message_protocol import OdSample
from tekleo_common_utils import UtilsImage


@injectable
class UtilsDatasetCoco:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage)):
        self.utils_image = utils_image

    def _build_default_class_labels(self, od_samples: List[OdSample]) -> List[str]:
        class_labels = []
        for s in od_samples:
            for i in s.items:
                class_labels.append(i.label)
        class_labels = set(class_labels)
        class_labels = list(class_labels)
        class_labels = sorted(class_labels)
        adjusted_class_labels = []
        adjusted_class_labels.append("__ignore__")
        adjusted_class_labels.extend(class_labels)
        return adjusted_class_labels

    # class_labels must have __ignore__ class as first item
    def save_samples_to_folder(self, od_samples: List[OdSample], folder_path: str, class_labels: List[str] = []) -> bool:
        # Fill class labels if default
        if class_labels is None or len(class_labels) == 0:
            class_labels = self._build_default_class_labels(od_samples)
        print('UtilsDatasetCoco.save_samples_to_folder(): class_labels=' + str(class_labels))

        # Make sure nested dirs exits
        if not os.path.exists(folder_path + "/JPEGImages"):
            os.makedirs(folder_path + "/JPEGImages")
        if not os.path.exists(folder_path + "/Visualization"):
            os.makedirs(folder_path + "/Visualization")

        data = dict(
            info=dict(
                description=None,
                url=None,
                version=None,
                year=datetime.datetime.now().year,
                contributor=None,
                date_created=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            ),
            licenses=[dict(url=None, id=0, name=None,)],
            images=[
                # license, url, file_name, height, width, date_captured, id
            ],
            type="instances",
            annotations=[
                # segmentation, area, iscrowd, image_id, bbox, category_id, id
            ],
            categories=[
                # supercategory, id, name
            ],
        )

        class_name_to_id = {}
        for i in range(0, len(class_labels)):
            class_id = i - 1  # starts with -1
            class_name = class_labels[i]
            if class_id == -1:
                assert class_name == "__ignore__"
                continue
            class_name_to_id[class_name] = class_id
            data["categories"].append(
                dict(supercategory=None, id=class_id, name=class_name,)
            )

        out_ann_file = folder_path + "/annotations.json"
        for od_sample_index in range(0, len(od_samples)):
            od_sample = od_samples[od_sample_index]
            image_width = od_sample.image.width
            image_height = od_sample.image.height
            print("UtilsDatasetCoco.save_samples_to_folder(): Generating coco dataset from:", od_sample.name)

            out_img_file = folder_path + "/JPEGImages/" + od_sample.name + ".jpg"
            self.utils_image.save_image_pil(od_sample.image, out_img_file)
            image_cv = self.utils_image.convert_image_pil_to_image_cv(od_sample.image)

            data["images"].append(
                dict(
                    license=0,
                    url=None,
                    file_name="JPEGImages/" + od_sample.name + ".jpg",
                    height=image_height,
                    width=image_width,
                    date_captured=None,
                    id=od_sample_index,
                )
            )

            masks = {}  # for area
            segmentations = collections.defaultdict(list)  # for segmentation
            for item in od_sample.items:
                points = [(p.x * image_width, p.y * image_height) for p in item.mask]
                label = item.label
                group_id = None
                shape_type = "polygon"
                mask = labelme.utils.shape_to_mask(
                    image_cv.shape[:2], points, shape_type
                )

                if group_id is None:
                    group_id = uuid.uuid1()

                instance = (label, group_id)

                if instance in masks:
                    masks[instance] = masks[instance] | mask
                else:
                    masks[instance] = mask

                if shape_type == "rectangle":
                    (x1, y1), (x2, y2) = points
                    x1, x2 = sorted([x1, x2])
                    y1, y2 = sorted([y1, y2])
                    points = [x1, y1, x2, y1, x2, y2, x1, y2]
                if shape_type == "circle":
                    (x1, y1), (x2, y2) = points
                    r = numpy.linalg.norm([x2 - x1, y2 - y1])
                    # r(1-cos(a/2))<x, a=2*pi/N => N>pi/arccos(1-x/r)
                    # x: tolerance of the gap between the arc and the line segment
                    n_points_circle = max(int(numpy.pi / numpy.arccos(1 - 1 / r)), 12)
                    i = numpy.arange(n_points_circle)
                    x = x1 + r * numpy.sin(2 * numpy.pi / n_points_circle * i)
                    y = y1 + r * numpy.cos(2 * numpy.pi / n_points_circle * i)
                    points = numpy.stack((x, y), axis=1).flatten().tolist()
                else:
                    points = numpy.asarray(points).flatten().tolist()

                segmentations[instance].append(points)
            segmentations = dict(segmentations)

            for instance, mask in masks.items():
                cls_name, group_id = instance
                if cls_name not in class_name_to_id:
                    continue
                cls_id = class_name_to_id[cls_name]

                mask = numpy.asfortranarray(mask.astype(numpy.uint8))
                mask = pycocotools.mask.encode(mask)
                area = float(pycocotools.mask.area(mask))
                bbox = pycocotools.mask.toBbox(mask).flatten().tolist()

                data["annotations"].append(
                    dict(
                        id=len(data["annotations"]),
                        image_id=od_sample_index,
                        category_id=cls_id,
                        segmentation=segmentations[instance],
                        area=area,
                        bbox=bbox,
                        iscrowd=0,
                    )
                )

            viz = image_cv
            if masks:
                labels, captions, masks = zip(
                    *[
                        (class_name_to_id[cnm], cnm, msk)
                        for (cnm, gid), msk in masks.items()
                        if cnm in class_name_to_id
                    ]
                )
                viz = imgviz.instances2rgb(
                    image=image_cv,
                    labels=labels,
                    masks=masks,
                    captions=captions,
                    font_size=15,
                    line_width=2,
                )
            out_viz_file = folder_path + "/Visualization/" + od_sample.name + ".jpg"
            self.utils_image.save_image_cv(viz, out_viz_file)

        with open(out_ann_file, "w") as f:
            json.dump(data, f)

        return True



import os
import torch
import numpy
import cv2
from numpy import ndarray
from typing import List
from detectron2 import model_zoo
from detectron2.config import get_cfg, CfgNode
from detectron2.engine import DefaultTrainer, DefaultPredictor
from injectable import injectable, autowired, Autowired
from tekleo_common_message_protocol import OdPrediction, RectanglePixel, PointPixel
from tekleo_common_utils import UtilsImage, UtilsOpencv


@injectable
class UtilsDetectronModel:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage), utils_opencv: Autowired(UtilsOpencv)):
        self.utils_image = utils_image
        self.utils_opencv = utils_opencv

    def is_gpu_available(self) -> bool:
        return torch.cuda.is_available()

    def build_config(
            self, model_zoo_config_name: str, number_of_classes: int,
            dataset_train_name: str, dataset_test_name: str,
            trained_model_output_dir: str,
            prediction_score_threshold: float,
            base_lr: float, max_iter: int, batch_size: int
    ) -> CfgNode:
        trained_model_weights_path = trained_model_output_dir + "/model_final.pth"

        cfg = get_cfg()
        cfg.merge_from_file(model_zoo.get_config_file(model_zoo_config_name))
        cfg.DATASETS.TRAIN = (dataset_train_name,)
        cfg.DATASETS.TEST = (dataset_test_name,)
        cfg.OUTPUT_DIR = trained_model_output_dir
        cfg.DATALOADER.NUM_WORKERS = 8

        # Force CPU
        if not self.is_gpu_available():
            cfg.MODEL.DEVICE = 'cpu'

        # Load previous version if present
        if os.path.exists(trained_model_weights_path):
            cfg.MODEL.WEIGHTS = trained_model_weights_path

        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = prediction_score_threshold
        cfg.SOLVER.IMS_PER_BATCH = 4
        cfg.SOLVER.BASE_LR = base_lr
        cfg.SOLVER.MAX_ITER = max_iter
        cfg.SOLVER.STEPS = []
        cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = batch_size
        cfg.MODEL.ROI_HEADS.NUM_CLASSES = number_of_classes
        cfg.TEST.DETECTIONS_PER_IMAGE = 100
        return cfg

    def build_trainer(self, cfg: CfgNode) -> DefaultTrainer:
        return DefaultTrainer(cfg)

    def build_predictor(self, cfg: CfgNode) -> DefaultPredictor:
        return DefaultPredictor(cfg)

    def _convert_outputs_build_region(self, box) -> RectanglePixel:
        x = box[0]
        y = box[1]
        w = box[2] - box[0]
        h = box[3] - box[1]
        return RectanglePixel(int(x), int(y), int(w), int(h))

    def _convert_outputs_build_mask(self, mask) -> List[PointPixel]:
        w, h = self.utils_opencv.get_dimensions_wh(mask)
        mask_background = numpy.zeros(shape=[h, w, 3], dtype=numpy.uint8)
        mask = numpy.where(mask == True, 255, mask_background)
        mask = mask.astype(numpy.uint8)
        mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
        contours, hierarchy = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        contours = sorted(contours, key = cv2.contourArea, reverse = True)
        largest_contour = contours[0]

        points = []
        for raw_point in largest_contour:
            x = raw_point[0][0]
            y = raw_point[0][1]
            points.append(PointPixel(x, y))
        return points

    def convert_outputs(self, class_labels: List[str], outputs) -> List[OdPrediction]:
        results = []
        instances = outputs["instances"].to("cpu")
        pred_boxes = instances.pred_boxes
        pred_masks = instances.pred_masks
        scores = instances.scores
        pred_classes = instances.pred_classes

        # Prepare mask arrays
        mask_array = pred_masks.numpy()
        mask_array = numpy.moveaxis(mask_array, 0, -1)

        # For each prediction
        for i in range(0, len(pred_boxes)):
            score = float(scores[i].numpy())
            label_key = int(pred_classes[i].numpy())
            label = class_labels[label_key]
            region = self._convert_outputs_build_region(pred_boxes[i].tensor.numpy()[0])
            mask = self._convert_outputs_build_mask(mask_array[:, :, i:(i+1)])
            prediction = OdPrediction(label, score, region, mask)
            results.append(prediction)

        return results

    def execute_predictor(self, predictor: DefaultPredictor, class_labels: List[str], image_cv: ndarray) -> List[OdPrediction]:
        outputs = predictor(image_cv)
        return self.convert_outputs(class_labels, outputs)

    def execute_trainer(self, trainer: DefaultTrainer):
        trainer.resume_or_load(resume=False)
        trainer.train()



# # Run dependency injections
import os
import tekleo_common_utils
from injectable import load_injection_container
load_injection_container(str(os.path.dirname(tekleo_common_utils.__file__)))
load_injection_container('../')
from tekleo_common_utils_ai.utils_dataset_labelme import UtilsDatasetLabelme
from tekleo_common_utils_ai.utils_dataset_coco import UtilsDatasetCoco

utils_dataset_labelme = UtilsDatasetLabelme()
utils_dataset_coco = UtilsDatasetCoco()

# Open labelme
labelme_folder_path = "dataset_labelme"
s1 = utils_dataset_labelme.load_sample_from_image_and_labelme_json(labelme_folder_path + "/signs1.jpg", labelme_folder_path + "/signs1.json")
print(s1)
s2 = utils_dataset_labelme.load_sample_from_folder("signs1", labelme_folder_path)
print(s2)
s3 = utils_dataset_labelme.load_samples_from_folder(labelme_folder_path)[0]
print(s3)

# Save to coco
coco_folder_path = "dataset_coco"
utils_dataset_coco.save_samples_to_folder([s3], coco_folder_path)
